-- Ron's call-by-need machine
module CBN (
  Term(..),  eval
) where 

import List
import Grammar
import FreshT

-- for an eventually fixed-up debug routine


class Show t => Thingie t where
    showThingie :: t -> F ()

instance Thingie Term where 
    showThingie t = 
        do
          lift $ putStr "\n\tTerm = "
          lift $ print t

instance Thingie Redex where
    showThingie r = 
        do
          case r of
            (Rterm a t) -> lift $ putStr "\n\tRedexTm = "
            (Rcont dc a) -> lift $ putStr "\n\tRedexDc = "
          lift $ print r

instance Thingie Answer where
    showThingie a = 
        do
          lift $ putStr "\n\tAnswer = "
          lift $ print a

----------------------------------------------------------------------------
-- Eval 

eval :: Bool -> Term -> IO ()
eval debug t = do a <- runF debug (refocus [] t)
                  putStr "Answer: " 
                  print t
                  putStr "Substituted: " 
                  print $ answer2term a

----------------------------------------------------------------------------
-- Reduce

sts2bodf (s,t) = BodF s t

reduce :: Context -> Redex -> F Answer
reduce ctxt r = do debug "reduce" ctxt r; reduce' r
    where
      reduce' (Rcont (Cont s1 ctxt2) a) =
          let v = answerValue a
              ctxta = answerCtx a
          in refocus (ctxt ++ ctxta ++ [BodF s1 v] ++ ctxt2) v
      reduce' (Rterm a t2) = 
          let v@(Lam s t1) = answerValue a
              ctxta = answerCtx a
          in do b <- isCaptured s
                Lam s' t1' <-  if b then freshLam v
                               else return v
                capture s'
                refocus (ctxt ++ ctxta ++ [BodF s' t2]) t1'

----------------------------------------------------------------------------
-- Refocus

-- The Term argument may contain free variables; they must covered by the set
-- of captured variables.

refocus :: Context -> Term -> F Answer
refocus ctxt t = do debug "refocus" ctxt t; refocus' t
  where 
    refocus' (Var s)   = 
      do assertCaptured s ("Free variable " ++ s ++ " not captured!")
         need ctxt [] s
    refocus' (Lam s t) = 
      rebuild ctxt (AValue s t)
    refocus' (App t1 t2) = 
      refocus (ctxt ++ [FunF t2]) t1

----------------------------------------------------------------------------
-- Rebuild

rebuild :: Context -> Answer -> F Answer
rebuild ctxt a = do debug "rebuild" ctxt a; rebuild' ctxt a
  where rebuild' [] a = return a
        rebuild' ctxt a = 
            let (ctxt',fr) = splitContext ctxt
            in case fr of 
                 FunF t' ->
                   reduce ctxt' (Rterm a t')
                 ArgF dc -> 
                   reduce ctxt' (Rcont dc a)
                 BodF s' t' -> -- capture free "s'" in a!
                   rebuild ctxt' (AFrame s' t' a)

----------------------------------------------------------------------------
-- Need

-- The second context NEVER captures the variable

need :: Context -> Context -> String -> F Answer
need ctxt1 ctxt2 s = 
  do debugNeed ctxt1 ctxt2 s
     let (ctxt',fr) = splitContext ctxt1 
       in case fr of 
          BodF s' t' | s == s' -> 
            let dc = Cont s ctxt2 -- captures s
            in refocus (ctxt' ++ [ArgF dc]) t' 
          fr -> need ctxt' (fr : ctxt2) s

----------------------------------------------------------------------------
-- Substitution
substTerm :: String -> String -> Term -> F Term
substTerm new s (Var s') = 
    if (s == s') then return (Var new)
    else return (Var s')
substTerm new s (App t1 t2) = 
    do t1' <- substTerm new s t1
       t2' <- substTerm new s t2
       return (App t1' t2')
substTerm new s (Lam s' t) = 
    if (s == s') then return (Lam s' t)
    else do Lam fs ft <- if (s' == new)
                         then freshLam (Lam s' t)
                         else return (Lam s' t)
            body <- substTerm new s ft
            return (Lam fs body)

freshLam :: Term -> F Term
freshLam (Lam s t) = 
    do s' <- newVar s
       t' <- substTerm s' s t
       return (Lam s' t')

----------------------------------------------------------------------------
-- Simple predicates and functions

splitContext :: Context -> (Context,Frame)
splitContext [] = error ("Unexpected empty context in splitContext\n")
splitContext fs = let (fs',[f]) = splitAt (length fs - 1) fs
                  in (fs',f)

isValue :: Term -> Bool
isValue (Lam _ _) = True
isValue _ = False



----------------------------------------------------------------------------
-- Assertions and debugging

slist :: [String] -> String
slist = concat . intersperse ", "

sctxt :: Context -> String
sctxt = concat . intersperse " +++ " . map show 

assertCaptured :: String -> String -> F ()
assertCaptured s msg = 
    do b <- isCaptured s
       if b 
          then return ()
          else error ("Assertion failed: " ++ msg ++ "\n")


debug :: (Thingie t) => String -> Context -> t -> F ()
debug s ctxt t = 
    do b <- debugFlag
       if (not b) then return ()
          else do lift $ putStr s
                  lift $ putStr "\n\tCaptured = "
                  ss <- allCaptured 
                  lift $ putStr ("{ " ++ (slist ss) ++ " }")
                  lift $ putStr "\n\tContext = "
                  lift $ putStr ("[ " ++ (sctxt ctxt) ++ " ]")
                  showThingie t
                  lift $ putStr "\n"
                  return ()

{--
debug' :: String -> Context -> Term -> F ()
debug' s ctxt t = 
    do b <- debugFlag
       if (not b) then return ()
          else do lift $ putStr s
                  lift $ putStr "\n\tCaptured = "
                  ss <- allCaptured 
                  lift $ putStr ("{ " ++ (slist ss) ++ " }")
                  lift $ putStr "\n\tContext = "
                  lift $ putStr ("[ " ++ (sctxt ctxt) ++ " ]")
                  lift $ putStr "\n\tTerm = "
                  lift $ print t
                  lift $ putStr "\n"
                  return ()
--}

debugNeed :: Context -> Context -> String -> F ()
debugNeed ctxt1 ctxt2 s = 
    do b <- debugFlag
       if (not b) then return ()
          else do lift $ putStr "need "
                  lift $ putStr s
                  lift $ putStr "\n\tCaptured = "
                  ss <- allCaptured 
                  lift $ putStr (" {" ++ (slist ss) ++ " }")
                  lift $ putStr "\n\tContext_1 = "
                  lift $ putStr ("[ " ++ (sctxt ctxt1) ++ " ]")
                  lift $ putStr "\n\tContext_2 = "
                  lift $ putStr ("[ " ++ (sctxt ctxt2) ++ " ]")
                  lift $ putStr "\n\n"
                  return ()

----------------------------------------------------------------------------

