-- Mostly a parser and prettyprinter: core copied from Lennart A.

module Grammar (
  Term(..), Cont(..), Redex(..), Answer(..), Frame(..), Context, fill,
  answer2term, answerValue, answerCtx
) where

import Maybe
import Data.Char(isAlphaNum)
import Text.PrettyPrint.HughesPJ (
  Doc, renderStyle, style, text, (<>), (<+>), parens, braces)
import Text.ParserCombinators.ReadP

----------------------------------------------------------------------------
-- Datatypes


-- t ::= x | \x.t | t t
data Term = Var String
          | Lam String Term
          | App Term Term

-- kx.E
data Cont = Cont String Context

-- r ::= kx.E a | a t
data Redex = Rcont Cont Answer
           | Rterm Answer Term

-- a ::= <E-, \x.t >  where E- ::= [(\x.[]) t]*
data Answer = AValue String Term
            | AFrame String Term Answer

-- f ::= [] t | (kx.E) [] | (\x.[]) t
data Frame = FunF Term
           | ArgF Cont
           | BodF String Term

-- E ::= [f]
type Context = [Frame]

----------------------------------------------------------------------------
-- Reading terms

instance Read Term where
    readsPrec _ = readP_to_S pLC

pLC, pLCAtom, pLCVar, pLCLam, pLCApp, pLCLet :: ReadP Term
pLC = pLCLam +++ pLCApp +++ pLCLet

pLCAtom = pLCVar +++ (do schar '('; e <- pLC; schar ')'; return e)

pLCVar = do v <- pVar
            return $ Var v

pLCLam = do schar '\\'
            v <- pVar
            schar '.'
            e <- pLC
            return $ Lam v e

pLCApp = do es <- many1 pLCAtom
            return $ foldl1 App es

pLCLet = let lcLet (x,e) b  = App (Lam x b) e
             pDef = do v <- pVar
                       schar '='
                       e <- pLC
                       return (v,e)
         in do sstring "let"
               bs <- sepBy pDef (schar ';')
               sstring "in"
               e <- pLC
               return $ foldr lcLet e bs

schar :: Char -> ReadP Char
schar c = do skipSpaces; char c

sstring :: String -> ReadP String
sstring c = do skipSpaces; string c

pVar :: ReadP String
pVar = do skipSpaces; munch1 isAlphaNum 


----------------------------------------------------------------------------
-- Substituting answers

answerCtx :: Answer -> Context
answerCtx (AValue s t) = []
answerCtx (AFrame s t a) = (BodF s t) : (answerCtx a)

answerValue :: Answer -> Term
answerValue (AValue s t) = Lam s t
answerValue (AFrame s t a) = answerValue a


answer2term :: Answer -> Term
answer2term a = answer2ev' [] a
    where
      answer2ev' :: [(String,Term)] -> Answer -> Term
      answer2ev' sts (AValue s t) = subst (Lam s t) sts
      answer2ev' sts (AFrame s t a) = answer2ev' ((s,t):sts) a

      subst :: Term -> [(String,Term)] -> Term
      subst (Var s) env =
          case lookup s env of
            Nothing -> Var s -- no match 
            Just t -> case t of
                        Var s' | s' == s -> (Var s) -- self-evaluating
                        _ -> subst t env
      subst (Lam s t) env = Lam s $ subst t $ (s,(Var s)):env
      subst (App t1 t2) env = App (subst t1 env) (subst t2 env)


----------------------------------------------------------------------------
-- Prettyprinting

instance Show Term where
    show = renderStyle style . ppLC 0

instance Show Answer where 
-- RG: use "fill" rather than answer2term for accurate rendering
    show = show . answer2term

instance Show Redex where
    show (Rcont cont ans) =
        renderStyle style $ ppCont 1 cont <+> ppLC 2 (answer2term ans)
    show (Rterm ans t) = show (App (answer2term ans) t)

fill :: Context -> Term -> Term
fill ctxt t = fill' (reverse ctxt) t
  where 
    fill' [] t = t
    fill' (FunF t' : fs) t = fill' fs (App t t')
    fill' (ArgF (Cont s e) : fs) t = fill' fs $ fill ([(BodF s t)] ++ e) (Var s)
    -- the following captures free s' in t
    fill' (BodF s' t' : fs) t = fill' fs (App (Lam s' t) t') 

ppLC :: Int -> Term -> Doc
ppLC _ (Var s) = text s
ppLC p (Lam s e) = pparens (p>0) $ text ("\\" ++ s ++ ".") <> ppLC 0 e
ppLC p (App f a) = pparens (p>1) $ ppLC 1 f <+> ppLC 2 a

pparens :: Bool -> Doc -> Doc
pparens True d = parens d
pparens False d = d

instance Show Frame where
    show = renderStyle style . ppFrame 0

ppFrame :: Int -> Frame -> Doc
ppFrame p (FunF t) = pparens (p>1) $ text "[]" <+> ppLC 2 t
ppFrame p (ArgF (Cont s e)) = pparens (p>1) $ ppMDC 1 s e <+> text "[]"
    where
      ppMDC p s c = pparens (p>0) $ 
                     text ("\\" ++ s ++ ".") <> 
                     (braces $ ppLC p (fill c (Var "[]"))) <>
                     text ("[" ++ s ++ "]")
ppFrame p (BodF s t) = pparens (p>1) $ 
                         (parens $ text ("\\" ++ s ++ ".[]")) <+> 
                         ppLC 2 t

ppCont :: Int -> Cont -> Doc
ppCont p (Cont s c) = pparens (p>0) $ 
                       text ("\\" ++ s ++ ".") <> 
                       (braces $ ppLC p (fill c (Var "[]"))) <> 
                       text ("[" ++ s ++ "]")
----------------------------------------------------------------------------
